/* 
 Trivial ping benchmark.
 Measures how many ping servers can run on same machine without affecting
 performance.

 On Red Hat Linux 7.2, I did

 ulimit -n 10000
 echo 30000 > /proc/sys/fs/file-max
 ulimit -p 5000
 echo 5000 > /proc/sys/kernel/threads-max

 and got up to 1150 children.  Increasing threads-max to 9000
 let me get up to 1300 children or so.

 Copyright Dan Kegel
 18 Feb 2002
 Licensed under the Gnu GPL v2 
 See http://www.kegel.com/dkcorbabench
*/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

static int die(int line)
{
	printf("Died at line %d!\n", line);
	kill(0, SIGKILL);
	exit(1);
}
#define DIE() die(__LINE__)

/* Display the system's memory statistics to stdout. */
static void	show_memory_usage(void)
{
	char buf[1024];
	int i;
	FILE *fp = fopen("/proc/meminfo", "r");
	if (!fp)
		return;
	
	for (i=0; i<2; i++) {
		if (!fgets(buf, sizeof(buf), fp))
			break;
		fputs(buf, stdout);
	}
	fclose(fp);
	fputs("\n", stdout);
}

struct childInfo {
	int pipe_to_child;
	int pipe_from_child;
};

/**
 Just read one byte from input pipe and echo it to output pipe forever
*/
static void respond_to_pings(struct childInfo *me)
{
	for (;;) {
		char buf[256];
		int nbytes;
		nbytes = read(me->pipe_to_child, buf, 1);
		if (nbytes != 1)
			DIE();
		nbytes = write(me->pipe_from_child, buf, 1);
		if (nbytes != 1)
			DIE();
	}
}

/**
 Ping n children, return how many milliseconds it takes.
*/
static int ping_n_children(int n, struct childInfo *children)
{
	int i;
	struct timeval start_timeval;
	struct timeval end_timeval;
	gettimeofday(&start_timeval, NULL);

	/* Do all the writes */
	for (i=0; i<n; i++) {
		char buf[1];
		int nbytes;
		buf[0] = (char)i;
		//printf("ping: i %d writing to fd %d\n", i, children[i].pipe_to_child);
		nbytes = write(children[i].pipe_to_child, buf, 1);
		if (nbytes != 1) {
			printf("nbytes %d errno %d writing 1 byte to fd %d\n", nbytes, errno, children[i].pipe_to_child);
			DIE();
		}
	}

	/* Do all the reads */
	for (i=0; i<n; i++) {
		char buf[1];
		int nbytes;
		buf[0] = 0;
		//printf("ping: i %d reading from fd %d\n", i, children[i].pipe_from_child);
		nbytes = read(children[i].pipe_from_child, buf, 1);
		if (nbytes != 1) {
			printf("nbytes %d errno %d reading 1 byte from fd %d\n", nbytes, errno, children[i].pipe_from_child);
			DIE();
		}
		if (buf[0] != (char)i)
			DIE();
	}
	gettimeofday(&end_timeval, NULL);
	return (end_timeval.tv_sec-start_timeval.tv_sec) * 1000
	     + (end_timeval.tv_usec-start_timeval.tv_usec) / 1000;
}

/**
 Fork 1 to max_n pingees.  
 Measure time to ping all children.
 Abort if latency exeeds max_t milliseconds.
*/

static int master_process(int max_n, int max_t)
{
	int i;
	struct childInfo *ci;
	ci = malloc(max_n * sizeof(*ci));

	for (i=0; i<max_n; i++) {
		int pipe_to_child[2];
		int pipe_from_child[2];
		int latency;
		if (pipe(pipe_to_child))
			DIE();
		//printf("1st pipe: %d %d\n", pipe_to_child[0], pipe_to_child[1]);
		if (pipe(pipe_from_child))
			DIE();
		//printf("2nd pipe: %d %d\n", pipe_from_child[0], pipe_from_child[1]);
		if (!fork()) {
			struct childInfo thischild;
			/* we're the child - start responding to pings */
			close(pipe_to_child[1]);
			close(pipe_from_child[0]);
			thischild.pipe_to_child = pipe_to_child[0];
			thischild.pipe_from_child = pipe_from_child[1];
			respond_to_pings(&thischild);
			exit(0);
		}
		close(pipe_to_child[0]);
		close(pipe_from_child[1]);
		ci[i].pipe_to_child = pipe_to_child[1];
		ci[i].pipe_from_child = pipe_from_child[0];

		if ((i % 50) == 0) {
			sleep(1);
			latency = ping_n_children(i+1, ci);
			printf("Latency to %d children is %d ms\n", i+1, latency);
			show_memory_usage();
			fflush(stdout);
			if (latency > max_t)
				break;
		}
	}
	return i;
}

int main(int argc, char **argv)
{
	int max_n;
	int max_t;
	int result;

	if (argc != 3) {
		fprintf(stderr, "usage: footprint0 max_n max_t\n");
		exit(1);
	}

	max_n = atoi(argv[1]);
	max_t = atoi(argv[2]);

	close(0);
	close(2);
	result = master_process(max_n, max_t);
	printf("Result: %d children supported at max_t of %d ms\n", result, max_t);

	return 0;
}
